# -*- coding: utf-8 -*-
import getpass

# different user has different log file to allow multiple user
logfile = '/tmp/fdi_tests_' + getpass.getuser() + '.log'
logdict = {
    "version": 1,
    "formatters": {
        "short": {
            "format": "%S %(funcName)s() %(message)s"
        },
        "full": {
            "format": "%(asctime)s %(name)s %(levelname)s %(args)s %(funcName)s():%(lineno)s - %(message)s",
            'datefmt': '%Y%m%d %H:%M:%S'
        }
    },
    "handlers": {
        "console": {
            "level": "INFO",
            "class": "logging.StreamHandler",
            "formatter": "full",
            "stream": "ext://sys.stdout"
        },
        "file": {
            "class": "logging.handlers.RotatingFileHandler",
            "level": "DEBUG",
            "formatter": "full",
            "filename": logfile,
            "maxBytes": 20000000,
            "backupCount": 3
        }
    },
    "loggers": {
        "": {
            "level": "INFO",
            "handlers": ["console", "file"],
            "propagate": False
        }
    },
    "root": {
        "level": "INFO",
        "handlers": ["console", "file"],
    },
    'disable_existing_loggers': False
}



# -*- coding: utf-8 -*-
from os.path import join
import logging
import getpass
import pwd

# logging level for server or possibly by client
pnsconfig = dict(logginglevel=logging.DEBUG)

# base url for webserver. Update version if needed.
pnsconfig['api_version'] = 'v0.6'
pnsconfig['baseurl'] = '/' + pnsconfig['api_version']

# base url for pool, you must have permission of this path, for example : /home/user/Documents
# this base pool path will be added at the beginning of your pool urn when you init a pool like:
# pstore = PoolManager.getPool('/demopool_user'), it will create a pool at /data.demopool_user/
# User can disable  basepoolpath by: pstore = PoolManager.getPool('/demopool_user', use_default_poolpath=False)
pnsconfig['base_poolpath'] = '/tmp'
pnsconfig['server_poolpath'] = '/tmp/data'  # For server
pnsconfig['defaultpool'] = 'pool_default'

dev = 1
if dev:
    # username, passwd, flask ip, flask port
    pnsconfig['node'] = {'username': 'foo',
                         'password': 'bar', 'host': '0.0.0.0', 'port': 5000}

    # server permission user
    pnsconfig['serveruser'] = 'mh'
    # PTS app permission user
    pnsconfig['ptsuser'] = 'mh'
    # on server
    home = '/cygdrive/c/Users/mh'
else:
    pnsconfig['node'] = {'username': 'luchangfa', 'password': '123456',
                         'host': '10.0.10.114', 'port': 9888}

    # server permission user
    pnsconfig['serveruser'] = 'apache'
    # PTS app permission user
    pnsconfig['ptsuser'] = 'pns'
    # on server
    home = '/root'

pnsconfig['auth_user'] = pnsconfig['node']['username']
pnsconfig['auth_pass'] = pnsconfig['node']['password']
pnsconfig['httphost'] = 'http://' + \
    pnsconfig['node']['host']+':'+str(pnsconfig['node']['port'])
pnsconfig['poolprefix'] = pnsconfig['httphost']
pnsconfig['mysql'] = {'host': 'ssa-mysql', 'port': 3306,
                      'user': 'root',  'password': '123456',
                      'database': 'users'}

# import user classes
# '/cygdrive/d/code/share/svom/products/projectclasses.py'
pnsconfig['userclasses'] = ''

phome = join(home, 'pns')
pnsconfig['paths'] = dict(
    pnshome=phome,
    inputdir=join(phome, 'input'),
    inputfiles=['pns.cat', 'pns.pn'],
    outputdir=join(phome, 'output'),
    outputfiles=['xycc.dat', 'atc.cc']
)

# the stateless data processing program that reads from inputdir and
# leave the output in the outputdir. The format is the input for subprocess()
h = pnsconfig['paths']['pnshome']
pnsconfig['scripts'] = dict(
    init=[join(h, 'initPTS'), ''],
    config=[join(h, 'configPTS'), ''],
    run=[join(h, 'runPTS'), ''],
    clean=[join(h, 'cleanPTS'), '']
)
del phome, h

# seconds
pnsconfig['timeout'] = 10



# -*- coding: utf-8 -*-
import logging
# create logger
logger = logging.getLogger(__name__)
#logger.debug('level %d' %  (logger.getEffectiveLevel()))


class Comparable(object):
    """
    """

    def __init__(self, **kwds):
        super(Comparable, self).__init__(**kwds)

    def compareTo(self, o):
        return self.__eq__(o)



# -*- coding: utf-8 -*-
""" https://stackoverflow.com/a/43573798/13472124 """
from .common import lls, trbk

import sys
import os.path

from importlib.abc import Loader, MetaPathFinder
from importlib.util import spec_from_file_location


import logging
# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))


class MyMetaFinder(MetaPathFinder):
    def find_spec(self, fullname, path, target=None):
        print('===', fullname, path, type(path))
        if path is None or path == "":
            path = [os.getcwd()]  # top level import --
        if "." in fullname:
            *parents, name = fullname.split(".")
        else:
            name = fullname
        print('***', name, path, type(path))
        for entry in path:
            if os.path.isdir(os.path.join(entry, name)):
                # this module has child modules
                filename = os.path.join(entry, name, "__init__.py")
                submodule_locations = [os.path.join(entry, name)]
            else:
                filename = os.path.join(entry, name + ".py")
                submodule_locations = None
            if not os.path.exists(filename):
                continue
            loader = MyLoader(filename)
            print('###', fullname, filename, submodule_locations, loader)
            ret = spec_from_file_location(fullname, filename,
                                          loader=loader,
                                          submodule_search_locations=submodule_locations)
            print('%%%', ret)
            return ret
        print('!!!', 'no loader', filename)
        return None  # we don't know how to import this


class SelectiveMetaFinder(MetaPathFinder):
    """ Raise ExcludedModule exception if loading modules whose names have any of cls.exclude.
    """
    class ExcludedModule(Exception):
        pass

    exclude = []

    def find_spec(self, fullname, path, target=None):
        # print('===', fullname, path, type(path))
        if path is None or path == "":
            path = [os.getcwd()]  # top level import --
        if "." in fullname:
            *parents, name = fullname.split(".")
        else:
            name = fullname
        #print('***', name, path, type(path), SelectiveMetaFinder.exclude)
        for x in SelectiveMetaFinder.exclude:
            if x == name:
                # print('!!!!!!!!!!!!!!!!!!!!!!!')
                raise(SelectiveMetaFinder.ExcludedModule(
                    name + ' is excluded from loading.'))
        for entry in path:
            if os.path.isdir(os.path.join(entry, name)):
                # this module has child modules
                filename = os.path.join(entry, name, "__init__.py")
                submodule_locations = [os.path.join(entry, name)]
            else:
                filename = os.path.join(entry, name + ".py")
                submodule_locations = None
            if not os.path.exists(filename):
                continue
            loader = MyLoader(filename)
            # print('###', fullname, filename, submodule_locations, loader)
            ret = spec_from_file_location(fullname, filename,
                                          loader=loader,
                                          submodule_search_locations=submodule_locations)
            #print('%%%', ret)
            return ret
        #print('!!!', 'no loader', filename)
        return None  # we don't know how to import this


class MyLoader(Loader):
    def __init__(self, filename):
        self.filename = filename
        #print('MMM ', filename)

    def create_module(self, spec):
        return None  # use default module creation semantics

    def exec_module(self, module):
        with open(self.filename, encoding='utf-8') as f:
            data = f.read()

        # manipulate data some way...
        #data += '\nb=42\n'
        logger.debug(self.filename)
        exec(data, vars(module))


def install():
    """Inserts the finder into the import machinery"""
    sys.meta_path.insert(0, MyMetaFinder())


def installSelectiveMetaFinder():
    """Inserts the finder into the import machinery"""
    sys.meta_path.insert(0, SelectiveMetaFinder())


def main(ipath='.'):
    if 0:
        install()

        import bar
        print(dir(bar))

    else:
        installSelectiveMetaFinder()
        SelectiveMetaFinder.exclude = ['notthis']
        # this is allowed
        from fdi.dataset.metadata import Parameter
        p = Parameter(0.99)
        assert p.value == 0.99
        # this is stopped.
        try:
            from fdi.dataset.notthis import BaseProduct
        except SelectiveMetaFinder.ExcludedModule as e:
            assert True
        else:
            assert False
        sys.path.insert(0, ipath)
        import bar
        assert hasattr(bar, 'k')


if __name__ == '__main__':

    main('tests')
    print('passed')



# -*- coding: utf-8 -*-
import logging
# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))


class Versionable(object):
    """
    for items being able to be defined with a Definition.
    """

    def __init__(self, **kwds):
        super(Versionable, self).__init__(**kwds)

    def getLastVersion(self, ref):
        """
        Returns the latest version of the given ProductRef.
        """

        return None

    def getVersions(self, ref):
        """
        Returns all the versions of the given ProductRef.
        """

        return []

    def saveProductRef(self, ref):
        """
        Saves the product referenced and returns the designated URN.
        """

        return None



# -*- coding: utf-8 -*-
from .odict import ODict
from .quantifiable import Quantifiable
from .eq import DeepEqual
from .copyable import Copyable
from .annotatable import Annotatable

from collections.abc import Container
import logging
# create logger
logger = logging.getLogger(__name__)
#logger.debug('level %d' %  (logger.getEffectiveLevel()))


class DataContainer(Annotatable, Quantifiable, Copyable, DeepEqual):
    """ A DataContainer is a composite of data and description.
    mh: note that There is no metadata.
    Implemented partly from AbstractDataWrapper.
    """

    def __init__(self, data=None, **kwds):
        """

        data: a Container. Default is None.
        """
        #print(__name__ + str(kwds))
        super(DataContainer, self).__init__(**kwds)

        if data is None or issubclass(data.__class__, Container):
            self.setData(data)
        else:
            raise TypeError('DataContainer needs a Container to initialize, not ' +
                            type(data).__name__)

    @property
    def data(self):
        return self.getData()

    @data.setter
    def data(self, newData):
        self.setData(newData)

    def setData(self, data):
        """ Replaces the current DataData with specified argument. 
        mh: subclasses can override this to add listener whenevery data is
        replaced
        """
        self._data = data

    def getData(self):
        """ Returns the data in this dw"""
        try:
            return self._data
        except AttributeError:
            od = ODict()
            self._data = od
            return od

    def hasData(self):
        """ Returns whether this data wrapper has data. """
        return self.getData() is not None and len(self.getData()) > 0

    def __repr__(self):
        return self.__class__.__name__ + \
            '{ description = "%s", data = "%s"}' % \
            (str(self.description), str(self.getData()))


class DataWrapper(DataContainer):
    """ A DataWrapper is a composite of data, unit and description.
    mh: note that all data are in the same unit. There is no metadata.
    Implemented from AbstractDataWrapper.
    """

    def __repr__(self):
        return self.__class__.__name__ + \
            '{ %s <%s>, description = "%s" }' % \
            (str(self.getData()), str(self.unit), str(self.description))


class DataWrapperMapper():
    """ Object holding a map of data wrappers. """

    def getDataWrappers(self):
        """ Gives the data wrappers, mapped by name. """
        return self._sets



# -*- coding: utf-8 -*-

from ..dataset.serializable import Serializable
from ..dataset.product import Product
from ..dataset.odict import ODict
import logging
# create logger
logger = logging.getLogger(__name__)
#logger.debug('level %d' % (logger.getEffectiveLevel()))


class ContextRuleException(ValueError):
    pass


class RefContainer(ODict):  # XXXXXXXX order
    """ A map where Rules of its owner Context are applied when put(k,v) is called, and the owner MapContext's ID can be put to v's parents list.

    Implemented.Composite so that RefContainer has a _STID when json.loads'ed.
    A MapContext has a _sets, which has a refs:RefContainer, which has a _sets, which has a name:ProductRef.
    when used as context.refs.get('x').product.description, the RefContainer is called with get() or __getitem__(), which calls superclass composite's _set's __getitem__()
    """

    def __init__(self, **kwds):
        """ 
        """
        super(RefContainer, self).__init__(**kwds)
        self.setOwner(None)

    @property
    def owner(self):
        """ Property """
        return self.getOwner()

    @owner.setter
    def owner(self, owner):
        """
        """
        self.setOwner(owner)

    def getOwner(self):
        """ Returns the reference container mapping
        """
        return self._owner

    def setOwner(self, owner):
        """ records who owns this container.

        Due to reason described in __setitem__ doc, existing refs will be set again upon owner set
        """

        if hasattr(self, '_owner') and self._owner is not None and id(self._owner) == id(owner):
            return
        self._owner = owner
        old = self.data
        self.data = dict()
        for k, r in old.items():
            self.__setitem__(k, r)
        del old

    def __setitem__(self, key, ref):
        """
            when deserialized contained refs will be set with _owner==None
            then the following are done when the assembled refContainer is
            set('refs',refC) to the owner context.
        """
        if key is None:
            raise KeyError()
        if ref is None:
            raise ValueError()

        # during deserialization __setitem__ is called before _owner is set by Context,setRefs()
        if hasattr(self, '_owner') and self._owner is not None:
            from .productref import ProductRef
            if isinstance(ref, ProductRef):
                ref.addParent(self._owner)
            else:
                raise TypeError(type(ref).__name__ +
                                ' is not a Product Reference')
            if not self._owner.applyRule(key, ref):
                ContextRuleException('Ref %s fails rule test of %s.' %
                                     (ref.toString(level=2),
                                      self._owner.__class__.__name__)
                                     )
        super().__setitem__(key, ref)

    def __delitem__(self, key):
        if key is None:
            raise KeyError()
        # during deserialization __setitem__ are called before _owner is set by Context,setRefs()
        if hasattr(self, '_owner'):
            from .productref import ProductRef
            ref = self.__getitem__(key)
            if isinstance(ref, ProductRef):
                ref.removeParent(self._owner)
        super().__delitem__(key)

    def clear(self):
        """ remove all productRefs """
        ks = list(self.keys())
        for k in ks:
            self.__delitem__(k)

    def put(self, key, ref):
        """ set label-ref pair after validating then add parent to the ref
        """
        self.__setitem__(key, ref)

    def set(self, key, ref):
        self.put(key, ref)

    def get(self, key):
        """
        """
        return self.__getitem__(key)

    def size(self):
        """ """
        return len(self.keys())

    # def __getstate__(self):
    #     """ Can be encoded with serializableEncoder """
    #     return ODict(
    #         data=self.data,
    #         _STID=self._STID)


class AbstractContext():
    """ A  special kind of Product that can hold references to other Products.

This abstract product introduces the lazy loading and saving of references to Products or ProductRefs that it is holding. It remembers its state.
http://herschel.esac.esa.int/hcss-doc-15.0/load/hcss_drm/api/herschel/ia/pal/Context.html
    """

    def __init__(self,  **kwds):
        """ Sets ``rule`` to ``None`` if ``zInfo`` does not have ``refs``, else to ``zInfo['refs']``.
        """
        super(AbstractContext, self).__init__(**kwds)

        self._dirty = False
        if 'refs' not in self.zInfo:
            self._rule = None  # None means there is no rule.
        else:
            self._rule = self.zInfo['refs']
        refC = None
        # this line must stay after _rule is set.
        self.setRefs(refC)

    @property
    def refs(self):
        """ Property """
        return self.getRefs()

    @refs.setter
    def refs(self, refs):
        """
        """
        self.setRefs(refs)

    def setRefs(self, refs):
        """ Changes/Adds the mapping container that holds references.
        """
        self.set('refs', refs)

    def set(self, name, refs):
        """ add owner to RefContainer
        """
        if isinstance(refs, RefContainer) and name == 'refs':
            refs.setOwner(self)
        super().set(name, refs)

    def getRefs(self):
        """ Returns the reference container mapping
        """
        return self.get('refs')

    def getAllRefs(self, recursive, includeContexts):
        """ Provides a set of the unique references stored in this context.
        """
        raise NotImplementedError()

    def hasDirtyReferences(self, storage):
        """ Returns a logical to specify whether this context has dirty references or not.
        """
        raise NotImplementedError()

    @classmethod
    def isContext(cls):
        """ Yields true if specified class belongs to the family of contexts.
        """
        return issubclass(cls, AbstractContext)

    def isValid(self):
        """ Provides a mechanism to ensure whether it is valid to store this context in its current state.
        """
        raise NotImplementedError()

    def readDataset(self, storage, table, defaultPoolId):
        """ Reads a dataset with information within this context that is normally not accessible from the normal Product interface."""
        raise NotImplementedError()

    def refsChanged(self):
        """ Indicates that the references have been changed in memory, which marks this context as dirty.
        """
        self._dirty = True

    def writeDataset(self,  storage):
        """ Creates a dataset with information within this context that is normally not accessible from the normal Product interface.
        """
        raise NotImplementedError()

    def getRule(self):
        """ Get the rule that controls the products to be added into the context.
        """
        return self._rule

    def setRule(self, rule):
        """ Set the rule that controls the products to be added into the context.
        """
        self._rule = rule

    def addRule(self, rule):
        """ Add to  the rule that controls the products to be added into the context.
         The new rule will be old rule AND added rule.
        """
        raise NotImplementedError()

    def applyRule(self, *args):
        """ returns True if the input ProductRef passes rule.

        If ``refs`` is not specified in ``zInfo`` return True.
        If ``zInfo['refs'] is empty, return True.

        Subclasses can override this method according to what is written in zInfo.
        """
        if self._rule is None:
            return True
        if len(self._rule) == 0:
            return True

        return True


class Context(AbstractContext, Product):
    """ See docstring of AbstractContext.
    """

    def __init__(self,  **kwds):
        """
        """
        super(Context, self).__init__(**kwds)

    def applyRule(self, *args):
        """ returns True if the input ProductRef passes rule.

        Default behavior is return the  superclass.applyRule() result.

        Subclasses can override this method according to what is written in zInfo.
        """
        return super().applyRule(*args)

    def getAllRefs(self, recursive=False, includeContexts=True, seen=None):
        """ Provides a set of the unique references stored in this context.
        This includes references that are contexts, but not the contents of these subcontexts. This is equivalent to getAllRefs(recursive=false, includeContexts= true).
        recursive - if true, include references in subcontexts
        includeContexts - if true, include references to contexts, not including this one
        """
        if not Context.isContext(self):
            raise TypeError('This ref does not point to a context')
        if seen is None:
            see = list()
        rs = list()
        for x in self.refs.values():
            if Context.isContext(x):
                if includeContexts:
                    if x == self:
                        pass
                    else:
                        if x not in rs:
                            rs.append(x)
                if recursive:
                    if x not in seen:
                        seen.append(x)
                        # enter the context
                        rs.add(x.getAllRefs(recursive=recursive,
                                            includeContexts=includeContexts,
                                            seen=seen))
                else:
                    pass
            else:
                # not contex
                # records
                if x not in rs:
                    rs.append(x)

        return rs

    def hasDirtyReferences(self, storage):
        """ Returns a logical to specify whether this context has dirty references or not.
        """
        return self._dirty

    @staticmethod
    def isContext(cls):
        """ Yields true if specified class belongs to the family of contexts.
        """
        return issubclass(cls, Context)


class AbstractMapContext(AbstractContext):
    """
    """

    def __init__(self,  **kwds):
        """
        """
        super(AbstractMapContext, self).__init__(**kwds)
        refC = RefContainer()
        # this line must stay after _rule is set.
        self.setRefs(refC)


class MapContext(Context):
    """ Allows grouping Products into a map of (String, ProductRef) pairs.
    New entries can be added if they comply to the adding rules of this context. The default behaviour is to allow adding any (String,ProductRef) given that ``rule`` is not set.

    An example::

        image     = ImageProduct(description="hi")
        spectrum  = SpectrumProduct(description="there")
        simple    = Product(description="everyone")

        context=MapContext()
        context.refs.put("x",ProductRef(image))
        context.refs.put("y",ProductRef(spectrum))
        context.refs.put("z",ProductRef(simple))
        print context.refs.size() # 3
        print context.refs.get('x').product.description # hi
        print context.refs.get('y').product.description # there
        print context.refs.get('z').product.description # everyone

    It is possible to insert a ProductRef at a specific key in the MapContext. The same insertion behaviour is followed as for a Python dict, in that if there is already an existing ProductRef for the given key, that ProductRef is replaced with the new one::

         product4=SpectrumProduct(description="everybody")
         context.refs.put("y", ProductRef(product4))
         product5=SpectrumProduct(description="here")
         context.refs.put("a", ProductRef(product5))

         print context.refs.get('x').product.description # hi
         print context.refs.get('y').product.description # everybody
         print context.refs.get('z').product.description # everyone
         print context.refs.get('a').product.description # here

    Note that the rules are only applied when putting an entry to the map!

    Be aware that

    1. the put() method of the map view may throw a ContextRuleException if the data added to the context violates the rules applied to the context.
    2. the put() method of the map view may throw a ValueError if either of the arguments to the put() method are null.


    BaseProduct--Product
                        \\
    AbstractContext------Contex---------MapContext

    """

    def __init__(self,  **kwds):
        """
        """
        super().__init__(**kwds)
        refC = RefContainer()
        # this line must stay after _rule is set.
        self.setRefs(refC)

    def addRule(self, rule):
        """ Add to  the rule that controls the products to be added into the context.
         The new rule will be old rule AND added rule.
        """
        self._rule.update(rule)

    def applyRule(self, key, ref):
        """ returns True if the input key name and ProductRef pass rule.

        Default behavior is return True if the fully qualified class name of the product which ref refers to equals ``zInfo['refs'][key]``, OR the superclass.applyRule() result.

        Subclasses can override this method according to what is written in zInfo.
        """
        if super().applyRule():
            return True
        if 'refs' in self.zInfo:
            return ref.urnobj.getTypeName() in self.zInfo['refs']



# -*- coding: utf-8 -*-

from os.path import join, expanduser, expandvars
import functools
import sys

import logging
# create logger
logger = logging.getLogger(__name__)
#logger.debug('logging level %d' % (logger.getEffectiveLevel()))


@functools.lru_cache(5)
def getConfig(conf='pns'):
    """ Imports a dict named [conf]config defined in ~/.config/[conf]local.py
    """
    # default configuration is provided. Copy pnsconfig.py to ~/.config/pnslocal.py
    env = expanduser(expandvars('$HOME'))
    # apache wsgi will return '$HOME' with no expansion
    env = '/root' if env == '$HOME' else env
    confp = join(env, '.config')
    sys.path.insert(0, confp)
    # logger.debug('Reading from configuration file in dir '+confp)

    try:
        c = __import__(conf+'local', globals(), locals(),
                       [conf+'config'], 0)
        logger.debug('Reading %s/%slocal.py done.' % (confp, conf))
        return c.__dict__[conf+'config']
    except ModuleNotFoundError as e:
        logger.warning(str(
            e) + '. Use default config in the package, such as fdi/pns/pnsconfig.py. Copy it to ~/.config/[package]local.py and make persistent customization there.')
        return {}



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#

import os
import sys
from datetime import datetime
import sphinx_rtd_theme
#import sphinx_readable_theme
#import sphinx_bootstrap_theme

sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

project = 'fdi'
year = datetime.now().year
copyright = '2019 - %d Maohai Huang, NAOC, ESA' % year
author = 'Maohai Huang'

# The full version, including alpha/beta/rc tags
# Version info -- read without importing
# https://github.com/aio-libs/aiohttp-theme/blob/master/setup.py
_locals = {}
with open('../../fdi/_version.py') as fp:
    exec(fp.read(), None, _locals)
release = _locals['__version__']


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ['sphinx.ext.autodoc']
extensions.append("sphinx_rtd_theme")
extensions.append('sphinx_copybutton')
extensions.append('sphinx.ext.napoleon')
extensions.append('sphinx.ext.inheritance_diagram')
extensions.append('sphinx.ext.viewcode')
# extensions.append(['IPython.sphinxext.ipython_console_highlighting',
#              'IPython.sphinxext.ipython_directive'])
# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', '*~']


# -- Options for HTML output -------------------------------------------------

# https://stackoverflow.com/a/32079202
def xsetup(app):
    # app.add_css_file("hatnotes.css")
    # app.add_css_file("custom.css")
    # app.add_css_file("bootstrap.css")
    # app.add_css_file("nature.css")
    app.add_css_file('fdi.css')


if 1:
    #from monokaimod import MonokaiMod
    pygments_style = 'fdi.utils.monokaimod.MonokaiMod'  # 'monokaimod'
else:
    pygments_style = 'monokai'

napoleon_google_docstring = False
napoleon_use_param = False
napoleon_use_ivar = True

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#

#html_theme = 'alabaster'
#html_theme = 'sphinxdoc'
html_theme = "sphinx_rtd_theme"
#html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
#html_theme = 'bootstrap'
#html_theme_path = sphinx_bootstrap_theme.get_html_theme_path()

# Alabaster side bar
xhtml_sidebars = {
    '**': [
        'about.html',
        'navigation.html',
        'relations.html',
        'searchbox.html',
        # 'donate.html',
    ]
}


# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# The following are from readthedocs.org
# https://docs.readthedocs.io/en/stable/guides/adding-custom-css.html

# These paths are either relative to html_static_path
# or fully qualified paths (eg. https://...)
html_css_files = [
    'css/custom.css',
    'css/fdi.css',
]

html_js_files = [
    'js/custom.js',
]

#html_style = 'css/yourtheme.css'

copybutton_prompt_text = r">>> |\.\.\. |\.\.\.: |\$ |In \[\d*\]: | {2,5}\.\.\.: | {5,8}: "
copybutton_prompt_is_regexp = True
copybutton_only_copy_prompt_lines = True
copybutton_image_path = 'copy-button-yellow.svg'
copybutton_remove_prompts = True

# Theme options are theme-specific and customize the look and feel of a
# theme further.
xhtml_theme_options = {
    'canonical_url': 'http://mercury.bao.ac.cn:9006/mh/fdi',
    'analytics_id': 'UA-XXXXXXX-1',  # Provided by Google in your dashboard
    'logo_only': False,
    'display_version': True,
    'prev_next_buttons_location': 'bottom',
    'style_external_links': False,
    'vcs_pageview_mode': '',
    'style_nav_header_background': 'white',
    # Toc options
    'collapse_navigation': True,
    'sticky_navigation': True,
    'navigation_depth': 4,
    'includehidden': True,
    'titles_only': False,
    'display_gitlab': True,
    'gitlab_url': 'http://mercury.bao.ac.cn:9006/mh/fdi',
}

XXhtml_theme_options = {
    'page_width': '1200px',
    'sidebar_width': '300px',
    'show_powered_by': True,
    'sidebar_collapse': True,
}


XXXhtml_theme_options = {
    # Navigation bar title. (Default: ``project`` value)
    'navbar_title': "fdi",

    # Tab name for entire site. (Default: "Site")
    'navbar_site_name': "ags",

    # A list of tuples containing pages or urls to link to.
    # Valid tuples should be in the following forms:
    #    (name, page)                 # a link to a page
    #    (name, "/aa/bb", 1)          # a link to an arbitrary relative url
    #    (name, "http://example.com", True) # arbitrary absolute url
    # Note the "1" or "True" value above as the third argument to indicate
    # an arbitrary url.
    'navbar_links': [
        ("Examples", "examples"),
        ("Link", "http://example.com", True),
    ],

    # Render the next and previous page links in navbar. (Default: true)
    'navbar_sidebarrel': True,

    # Render the current pages TOC in the navbar. (Default: true)
    'navbar_pagenav': True,

    # Tab name for the current pages TOC. (Default: "Page")
    'navbar_pagenav_name': "Page",

    # Global TOC depth for "site" navbar tab. (Default: 1)
    # Switching to -1 shows all levels.
    'globaltoc_depth': 2,

    # Include hidden TOCs in Site navbar?
    #
    # Note: If this is "false", you cannot have mixed ``:hidden:`` and
    # non-hidden ``toctree`` directives in the same page, or else the build
    # will break.
    #
    # Values: "true" (default) or "false"
    'globaltoc_includehidden': "true",

    # HTML navbar class (Default: "navbar") to attach to <div> element.
    # For black navbar, do "navbar navbar-inverse"
    'navbar_class': "navbar navbar-inverse",

    # Fix navigation bar to top of page?
    # Values: "true" (default) or "false"
    'navbar_fixed_top': "true",

    # Location of link to source.
    # Options are "nav" (default), "footer" or anything else to exclude.
    'source_link_position': "nav",
}
xxxhtml_theme_options = {
    # Bootswatch (http://bootswatch.com/) theme.
    #
    # Options are nothing (default) or the name of a valid theme
    # such as "cosmo" or "sandstone".
    #
    # The set of valid themes depend on the version of Bootstrap
    # that's used (the next config option).
    #
    # Currently, the supported themes are:
    # - Bootstrap 2: https://bootswatch.com/2
    # - Bootstrap 3: https://bootswatch.com/3
    'bootswatch_theme': "superhero",

    # Choose Bootstrap version.
    # Values: "3" (default) or "2" (in quotes)
    'bootstrap_version': "3",
}



# -*- coding: utf-8 -*-

from collections.abc import Mapping
from types import MappingProxyType
import copy
from immutabledict import immutabledict


def make_readonly(d, excluded=None):
    """ Recursively make values of Mapping type at all levels read-only.
    """
    if excluded is None:
        excluded = set()
    else:
        i = id(d)
        if i in excluded:
            return d
        else:
            excluded.add(i)
    for k, v in d.items():
        if issubclass(v.__class__, Mapping):
            d[k] = make_readonly(v, excluded=excluded)
        else:
            continue
    return d


class frozendict(dict):
    """
    Extended version of frozent dict.

    This extended version of the code above will also "freeze" dictionaries and lists stored as values of the dictionary. If dictionaries/lists are found among the values, they will be handled recursively.

    Oren Tirosh, Ero Carrera, Søren Løvborg

    https://code.activestate.com/recipes/414283/
    """
    def _blocked_attribute(obj):
        raise AttributeError("A frozendict cannot be modified.")
    _blocked_attribute = property(_blocked_attribute)

    __delitem__ = __setitem__ = clear = _blocked_attribute
    pop = popitem = setdefault = update = _blocked_attribute

    def __new__(cls, *args, **kw):
        new = dict.__new__(cls)

        args_ = []
        for arg in args:
            if isinstance(arg, dict):
                arg = copy.copy(arg)
                for k, v in arg.items():
                    if isinstance(v, dict):
                        arg[k] = frozendict(v)
                    elif isinstance(v, list):
                        v_ = list()
                        for elm in v:
                            if isinstance(elm, dict):
                                v_.append(frozendict(elm))
                            else:
                                v_.append(elm)
                        arg[k] = tuple(v_)
                args_.append(arg)
            else:
                args_.append(arg)

        dict.__init__(new, *args_, **kw)
        return new

    def __init__(self, *args, **kw):
        pass

    def __hash__(self):
        try:
            return self._cached_hash
        except AttributeError:
            h = self._cached_hash = hash(frozenset(self.items()))
            return h

    def __repr__(self):
        return "frozendict(%s)" % dict.__repr__(self)


ReadOnlyDict = frozendict


class aReadOnlyDict(immutabledict):

    def __getitem_a_(self, key):
        it = super().__getitem__(key)

        if issubclass(it.__class__, dict):
            rodict = ReadOnlyDict(it)
            return rodict
        return it


class bReadOnlyDict(dict):
    """
    Dictionary that is read-only and returns all values of dict type to ReadOnlyDict type so they, too, cannot be modified.

    Based on https://stackoverflow.com/a/61700719/13472124


    """

    # def __init__(self, d):

    def __readonly__(self, *args, **kwargs):
        raise RuntimeError("Cannot modify ReadOnlyDict")

    __setitem__ = __readonly__
    __delitem__ = __readonly__
    pop = __readonly__
    popitem = __readonly__
    clear = __readonly__
    update = __readonly__
    setdefault = __readonly__
    __copy__ = __readonly__  # dict.copy
    __deepcopy__ = __readonly__  # copy._deepcopy_dispatch.get(dict)
    del __readonly__

    def __getitem__(self, key):
        it = super().__getitem__(key)
        # print(key)
        import pdb
        pdb.set_trace()

        if issubclass(it.__class__, dict):
            rodict = ReadOnlyDict(it)
            return rodict
        return it



# -*- coding: utf-8 -*-
import logging
import json
import codecs
import builtins
from collections import UserDict
from collections.abc import MutableMapping as MM, MutableSequence as MS, MutableSet as MS

from .odict import ODict
from .classes import Classes
from ..utils.common import lls

import sys
if sys.version_info[0] >= 3:  # + 0.1 * sys.version_info[1] >= 3.3:
    PY3 = True
    strset = (str, bytes)
else:
    PY3 = False
    strset = (str, unicode)

# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))


''' Note: this has to be in a different file where other interface
classes are defined to avoid circular dependency (such as ,
Serializable.
'''

BD = builtins.__dict__


def constructSerializable(obj, lgb=None, debug=False):
    """ mh: reconstruct object from the output of jason.loads().
    Recursively goes into nested class instances that are not
    encoded by default by JSONEncoder, instantiate and fill in
    variables.
    Objects to be deserialized must have their classes loaded.
    _STID cannot have module names in it (e.g.  dataset.Product)
    or locals()[classname] or globals()[classname] will not work. See alternative in
    https://stackoverflow.com/questions/452969/does-python-have-an-equivalent-to-java-class-forname

    """
    global indent
    indent += 1
    spaces = '  ' * indent

    classname = obj.__class__.__name__
    if debug:
        print(spaces + '===OBJECT %s ===' % lls(obj, 150))
    if not hasattr(obj, '__iter__') or issubclass(obj.__class__, strset):
        if debug:
            print(spaces + 'Find non-iter <%s>' % classname)
        indent -= 1
        return obj

    # process list first
    if isinstance(obj, list):
        if debug:
            print(spaces + 'Find list <%s>' % classname)
        inst = []
        # loop i to preserve order
        for i in range(len(obj)):
            x = obj[i]
            xc = x.__class__
            if debug:
                print(spaces + 'looping through list %d <%s>' %
                      (i, xc.__name__))
                if issubclass(xc, (list, dict)):
                    des = constructSerializable(x, lgb=lgb, debug=debug)
            else:
                des = x
            inst.append(des)
        if debug:
            print(spaces + 'Done with list <%s>' % (classname))
        indent -= 1
        return inst

    if not '_STID' in obj:
        """ This object is supported by JSON encoder """
        if debug:
            print(spaces + 'Find non-_STID. <%s>' % classname)
        inst = obj
    else:
        classname = obj['_STID']
        if debug:
            print(spaces + 'Find _STID <%s>' % classname)
        # process types wrapped in a dict
        if PY3 and classname == 'bytes':
            inst = codecs.decode(obj, 'hex')
            if debug:
                print(spaces + 'Instanciate hex')
            indent -= 1
            return inst
        if classname in lgb:
            # Now we have a blank instance.
            inst = lgb[classname]()
            if debug:
                print(spaces + 'Instanciate custom obj <%s>' % classname)
        elif classname == 'ellipsis':
            if debug:
                print(spaces + 'Instanciate Ellipsis')
            indent -= 1
            return Ellipsis
        elif classname in BD and 'obj' in obj:
            o = constructSerializable(obj['obj'], lgb=lgb, debug=debug)
            inst = BD[classname](o)
            if debug:
                print(spaces + 'Instanciate builtin %s' % obj['obj'])
            indent -= 1
            return inst
        else:
            raise ValueError('Class %s is not known.' % classname)
    if debug:
        print(spaces + 'Go through properties of instance')
    for (k, v) in obj.items():
        """ loop through all key-value pairs. """
        if k == '_STID':
            continue
        # deserialize v
        # should be object_pairs_hook in the following if... line
        if issubclass(v.__class__, (dict, list)):
            if debug:
                print(spaces + '[%s]value(dict/list) <%s>: %s' %
                      (k, v.__class__.__qualname__,
                       lls(list(iter(v)), 70)))
            desv = constructSerializable(v, lgb=lgb, debug=debug)
        else:
            if debug:
                print(spaces + '[%s]value(simple) <%s>: %s' %
                      (k, v.__class__.__name__, lls(v, 70)))
            if 1:
                desv = v
            else:
                if isinstance(v, str) or isinstance(v, bytes):
                    try:
                        desv = int(v)
                    except ValueError:
                        desv = v

        # set k with desv
        if issubclass(inst.__class__, (MM)):    # should be object_pairs_hook
            inst[k] = desv
            if debug:
                print(spaces + 'Set dict/usrd <%s>[%s] = %s <%s>' %
                      ((inst.__class__.__name__), str(k), lls(desv, 70), (desv.__class__.__name__)))
        else:
            setattr(inst, k, desv)
            if debug:
                print(spaces + 'set non-dict <%s>.%s = %s <%s>' %
                      ((inst.__class__.__name__), str(k), lls(desv, 70), (desv.__class__.__name__)))
    indent -= 1
    return inst


class IntDecoder(json.JSONDecoder):
    """ adapted from https://stackoverflow.com/questions/45068797/how-to-convert-string-int-json-into-real-int-with-json-loads
    modified to also convert keys in dictionaries.
    """

    def decode(self, s):
        # result = super(Decoder, self).decode(s) for Python 2.x
        result = super(IntDecoder, self).decode(s)
        return self._decode(result)

    def _decode(self, o):
        if isinstance(o, str) or isinstance(o, bytes):
            try:
                return int(o)
            except ValueError:
                return o
        elif isinstance(o, dict):
            return dict({self._decode(k): self._decode(v) for k, v in o.items()})
        elif isinstance(o, list):
            return [self._decode(v) for v in o]
        else:
            return o


class IntDecoderOD(IntDecoder):
    """ Uses ODict
    """

    def _decode(self, o):
        if isinstance(o, str) or isinstance(o, bytes):
            try:
                return int(o)
            except ValueError:
                return o
        elif isinstance(o, dict):
            return ODict({self._decode(k): self._decode(v) for k, v in o.items()})
        elif isinstance(o, list):
            return [self._decode(v) for v in o]
        else:
            return o


def deserialize(js, lgb=None, debug=False, usedict=True):
    """ Loads classes with _STID from the results of serialize.

    if usedict is True dict insted of ODict will be used.
    """

    if lgb is None:
        lgb = Classes.mapping

    if not isinstance(js, strset) or len(js) == 0:
        return None
    # debug = False  # True if issubclass(obj.__class__, list) else False
    try:
        if usedict:
            obj = json.loads(js)  # , cls=IntDecoder)
        else:
            # , cls=IntDecoderOD)
            obj = json.loads(js, object_pairs_hook=ODict)
    except json.decoder.JSONDecodeError as e:
        msg = '\nBad string to decode as JSON=====>\n%s\n<======\n' %\
            lls(js, 500)
        logging.error(msg)
        obj = msg
    if debug:
        # print('load-str ' + str(o) + ' class ' + str(o.__class__))
        print('-------- json loads returns: --------\n' + str(obj))

    global indent
    indent = -1
    return constructSerializable(obj, lgb=lgb, debug=debug)



# -*- coding: utf-8 -*-

from .context import Context
from .urn import Urn
from .comparable import Comparable
from ..dataset.product import Product
from ..dataset.odict import ODict
from ..dataset.serializable import Serializable
from ..dataset.attributable import Attributable
from collections import OrderedDict
import logging
# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))


class ProductRef(Attributable, Serializable, Comparable):
    """ A lightweight reference to a product that is stored in a ProductPool or in memory.
    """

    def __init__(self, urn=None, poolname=None, product=None, meta=None, **kwds):
        """ Urn can be the string or URNobject. if product is provided create an in-memory URN.
        Poolname if given overrides the pool name in urn, and causes metadata to be loaded from pool, unless this prodref points to a mempool.
        If meta is given, it will be used instead of that from poolname.
        A productref created from a single product will result in a memory pool urn, and the metadata won't be loaded.
        """
        super(ProductRef, self).__init__(**kwds)
        if issubclass(urn.__class__, str):
            urnobj = Urn(urn)
        elif issubclass(urn.__class__, Urn):
            urnobj = urn
        elif issubclass(urn.__class__, Product):
            # allow ProductRef(p) where p is a Product
            if product is None:
                product = urn
        else:
            urnobj = None

        if product:
            from .poolmanager import PoolManager, DEFAULT_MEM_POOL
            from . import productstorage
            pool = PoolManager.getPool(poolurl='mem:///' + DEFAULT_MEM_POOL)
            st = productstorage.ProductStorage(pool)
            urnobj = st.save(product, geturnobjs=True)
            # a lone product passed to prodref will be stored to mempool

        self.setUrnObj(urnobj, poolname, meta)

        if product and isinstance(product, Context):
            self._product = product

        self._parents = []

    @property
    def product(self):
        return self.getProduct()

    def getProduct(self):
        """ Get the product that this reference points to.

        If the product is a Context, it is kept internally, so further accesses don't need to ask the storage for loading it again.
        Otherwise, the product is returned but the internal reference remains null, so every call to this method involves a request to the storage.

        This way, heavy products are not kept in memory after calling this method, thus maintaining the ProductRef a lighweight reference to the target product.

        In case of a Context, if it is wanted to free the reference, call unload().

        Returns:
        the product
        """

        if self._poolname is None:
            return None
        if hasattr(self, '_product') and self._product is not None:
            return self._product
        from .poolmanager import PoolManager
        p = PoolManager.getPool(self._poolname).loadProduct(self.getUrn())
        if issubclass(p.__class__, Context):
            self._product = p
        return p

    def getPoolname(self):
        """ Returns the name of the product pool associated.
        """
        return self._poolname

    def getStorage(self):
        """ Returns the product storage associated.
        """
        raise NotImplementedError
        return self._storage

    def setStorage(self, storage):
        """ Sets the product storage associated.
        """
        raise NotImplementedError
        self._storage = storage
        # if hasattr(self, '_urn') and self._urn:
        #    self._meta = self._storage.getMeta(self._urn)

    def getType(self):
        """ Specifies the Product class to which this Product reference is pointing to.
        """
        return self._urnobj.getType()

    @property
    def urn(self):
        """ Property """
        return self.getUrn()

    @urn.setter
    def urn(self, urn):
        """
        """
        self.setUrn(urn)

    def setUrn(self, urn):
        """
        """
        self.setUrnObj(Urn(urn))

    def getUrn(self):
        """ Returns the Uniform Resource Name (URN) of the product.
        """
        return self._urnobj.urn

    @property
    def urnobj(self):
        """ Property """
        return self.getUrnObj()

    @urnobj.setter
    def urnobj(self, urnobj):
        """
        """
        self.setUrnObj(urnobj)

    def setUrnObj(self, urnobj, poolname=None, meta=None):
        """ sets urn

        A productref created from a single product will result in a memory pool urn, and the metadata won't be loaded.

        Parameters:
        -----------
        urnobj: Urn
        a URN object.
        poolname: str
        if given overrides the pool name in urn, and causes metadata to be loaded from pool.
        meta: MetaData
        If  is given, it will be used instead of that from poolname.
        """
        if urnobj is not None:
            # logger.debug(urnobj)
            assert issubclass(urnobj.__class__, Urn)

        self._urnobj = urnobj
        if urnobj is not None:
            self._urn = urnobj.urn

            from .poolmanager import PoolManager, DEFAULT_MEM_POOL
            from . import productstorage
            loadmeta = (poolname or meta) and poolname != DEFAULT_MEM_POOL
            if poolname is None:
                poolname = urnobj.pool
            pool = PoolManager.getPool(poolname)
            self._meta = (meta if meta else pool.meta(
                urnobj.urn)) if loadmeta else None
            self._poolname = poolname
            self._product = None
        else:
            self._urn = None
            self._poolname = None
            self._meta = None
            self._product = None

    def getUrnObj(self):
        """ Returns the URN as an object.
        """
        return self._urnobj

    @property
    def meta(self):
        """ Property """
        return self.getMeta()

    def getMeta(self):
        """ Returns the metadata of the product.
        """
        return self._meta

    def getHash(self):
        """ Returns a code number for the product; actually its MD5 signature. 
        This allows checking whether a product already exists in a pool or not.
        """
        raise NotImplementedError()

    def getSize(self):
        """ Returns the estimated size(in bytes) of the product in memory. 
        Useful for providing this information for a user that wants to download the product from a remote site.
        Returns:
        the size in bytes
        """
        raise NotImplementedError()

    def unload(self):
        """ Clear the cached meta and frees internal reference to the product, so it can be garbage collected.
        """
        self._product = None
        self._meta = None

    def isLoaded(self):
        """ Informs whether the pointed product is already loaded.
        """
        return self._product is not None

    def addParent(self, parent):
        """ add a parent
        """
        if any(id(parent) == id(x) for x in self._parents):
            return
        self._parents.append(parent)

    def removeParent(self, parent):
        """ remove a parent

        :param parent: 

        """
        if parent:
            self._parents.remove(parent)

    @property
    def parents(self):
        """ property """

        return self.getParents()

    @parents.setter
    def parents(self, parents):
        """ property """

        self.setParents(parents)

    def getParents(self):
        """ Return the in-memory parent context products of this reference.
        That is, the contexts in program memory that contain this product reference object. 
        A context that contains a different product reference object pointing to the same URN is not a parent of this product reference.

        Furthermore, it should be understood that this method does not return the parent contexts of the product pointed to by this reference as stored in any underlying pool or storage.

        Returns:
        the parents 
        """
        return self._parents

    def setParents(self, parents):
        """ Sets the in-memory parent context products of this reference.

        :param parents: 

        """
        self._parents = parents

    def equals(self, o, verbose=False):
        """     true if o is a non-null ProductRef, with the same Product type than this one, and:

        urns and products are null in both refs, or
        nurs are equal and products are null, or # <-- mh
        urns are null in both refs, and their products are equal, or
        urns and products are equal in both refs
        """
        t1 = issubclass(o.__class__, ProductRef)
        if not t1:
            if verbose:
                msg = 'Input o is not a ProductRef'
                return msg
            return False
        if self._product is None:
            if o._product is None:
                if o._urnobj is None and self._urnobj is None or o._urnobj == self._urnobj:
                    if verbose:
                        msg = 'Both onject._products are None or have equal URN.'
                        return msg
                    return True
            else:
                if verbose:
                    msg = 'Self._product is None but not for the other obj.'
                    return msg
                return False
        else:
            if self._product == o._product and (self._product.type == o._product.type):
                if (o._urnobj is None and self._urnobj is None) or \
                   (o._urnobj == self._urnobj):
                    if verbose:
                        print('True due to equal _project and _urnobj')
                    return True

        return False

    def __eq__(self, o):
        """ has the same Urn.
        """
        return self.equals(o)

    def __hash__(self):
        """ returns hash of the URN object
        """
        return hash(self._urnobj)

    def __repr__(self):
        return self.toString(level=1)

    def toString(self, level=0, **kwds):
        """
        """
        s = self.__class__.__name__
        s += ' {' + self.urn
        if level == 0:
            s += '\n# Parents=' + \
                str([str(id(p)) + ' ' + p.__class__.__name__ +
                     '"' + p.description + '"'
                     for p in self.parents]) + '\n'
            s += '# meta=' + self.getMeta().toString(level=level, **kwds)
        else:
            s += ' Parents=' + str([id(p) for p in self.parents])
            s += ' meta= ' + 'None' if self.getMeta() is None else self.getMeta().toString(level=level, **kwds)
        s += '}'
        return s

    def __getstate__(self):
        """ Can be encoded with serializableEncoder """
        return OrderedDict(urnobj=self.urnobj if issubclass(self.urnobj.__class__, Urn) else None,
                           _meta=self.getMeta(),
                           _STID=self._STID)



# -*- coding: utf-8 -*-
import pytest
from fdi.dataset.serializable import serialize
from fdi.dataset.deserialize import deserialize
from fdi.dataset.product import Product
from fdi.pal.poolmanager import PoolManager
from fdi.utils.getconfig import getConfig
from fdi.utils.common import lls, trbk
#from fdi.pns import httppool_server as HS


import filelock
import sys
import base64
from urllib.request import pathname2url
from requests.auth import HTTPBasicAuth
import requests
import random
import os
import time
from collections.abc import Mapping

import asyncio
import aiohttp

from fdi.pns.jsonio import getJsonObj, postJsonObj, putJsonObj, commonheaders
from fdi.utils.options import opt

# default configuration is provided. Copy pnsconfig.py to ~/local.py
from fdi.pns.pnsconfig import pnsconfig as pc


def setuplogging():
    import logging
    import logging.config
    from . import logdict

    # create logger
    logging.config.dictConfig(logdict.logdict)
    logging.getLogger("requests").setLevel(logging.WARN)
    logging.getLogger("urllib3").setLevel(logging.WARN)
    logging.getLogger("filelock").setLevel(logging.WARN)
    return logging


logging = setuplogging()
logger = logging.getLogger()


pc.update(getConfig())
logger.setLevel(logging.INFO)
logger.debug('logging level %d' % (logger.getEffectiveLevel()))


if 0:
    @pytest.fixture(scope="module")
    def runserver():
        from fdi.pns.httppool_server import app
        app.run(host='127.0.0.1', port=5000,
                threaded=False, debug=verbose, processes=5)

        return smtplib.SMTP("smtp.gmail.com", 587, timeout=5)


testname = 'SVOM'
aburl = 'http://' + pc['node']['host'] + ':' + \
    str(pc['node']['port']) + pc['baseurl']

up = bytes((pc['node']['username'] + ':' +
            pc['node']['password']).encode('ascii'))
code = base64.b64encode(up).decode("ascii")
commonheaders.update({'Authorization': 'Basic %s' % (code)})
del up, code

# last timestamp/lastUpdate
lupd = 0

auth_user = pc['auth_user']
auth_pass = pc['auth_pass']
test_poolid = 'test'
basepath = pc['server_poolpath']
prodt = 'fdi.dataset.product.Product'

# http server pool
schm = 'server'
basepath = PoolManager.PlacePaths[schm]
poolpath = os.path.join(basepath, pc['api_version'])
# client side
cschm = 'http'
api_baseurl = cschm + '://' + \
    PoolManager.PlacePaths[cschm] + '/'


if 0:
    poststr = 'curl -i -H "Content-Type: application/json" -X POST --data @%s http://localhost:5000%s --user %s'
    cmd = poststr % ('resource/' + 'nodetestinput.jsn',
                     pathname2url(pc['baseurl'] + '/' +
                                  nodetestinput['creator'] + '/' +
                                  nodetestinput['rootcause']),
                     'foo:bar')
    print(cmd)
    os.system(cmd)
    sys.exit()


def checkserver():
    """ make sure the server is running when tests start
    """
    global testpns
    # check if data already exists
    o = getJsonObj(aburl + '/')
    assert o is not None, 'Cannot connect to the server'
    logger.info('initial server response %s' % (str(o)[:100] + '...'))
    # assert 'result' is not None, 'please start the server to refresh.'
    # initialize test data.


def issane(o):
    """ basic check on return """
    global lupd
    assert o is not None, "Server is having trouble"
    assert 'error' not in o, o['error']
    assert o['timestamp'] > lupd
    lupd = o['timestamp']


def check0result(result, msg):
    # if msg is string, an exception must have happened
    assert result == 0, 'Error %d testing script "run". msg: ' + str(msg)
    assert msg == '' or not isinstance(msg, (str, bytes)), msg


def est_getpnspoolconfig():
    ''' gets and compares pnspoolconfig remote and local
    '''
    logger.info('get pnsconfig')
    o = getJsonObj(aburl + '/pnsconfig')
    issane(o)
    r = o['result']
    # , deepcmp(r['scripts'], pc['scripts'])
    assert r['scripts'] == pc['scripts']
    return r


def checkContents(cmd, filename):
    """ checks a GET commands return matches contents of a file.
    """
    o = getJsonObj(aburl + cmd)
    issane(o)
    with open(filename, 'r') as f:
        result = f.read()
    assert result == o['result'], o['message']


# TEST HTTPPOOL  API


def check_response(o, failed_case=False):
    global lupd
    assert o is not None, "Server is having trouble"
    if not failed_case:
        assert 'FAILED' != o['result'], o['result']
        assert o['timestamp'] > lupd
        lupd = o['timestamp']
    else:
        assert 'FAILED' == o['result'], o['result']


def clear_server_poolpath(poolid):
    """ deletes files in the given poolid in server pool dir. """
    logger.info('clear server pool dir ' + poolid)
    path = os.path.join(poolpath, poolid)
    if os.path.exists(path):
        if path == '/':
            raise ValueError('!!!!! Cannot delete root.!!!!!!!')
        else:
            os.system('rm -rf ' + path)
        # x = Product(description='desc test case')
        # x.creator = 'test'
        # data = serialize(x)
        # url = api_baseurl + test_poolid + '/fdi.dataset.product.Product/0'
        # x = requests.post(url, auth=HTTPBasicAuth(auth_user, auth_pass), data=data)


def get_files(poolid):
    """ returns a list of files in the given poolid in server pool dir. """

    path = os.path.join(poolpath, poolid)
    if os.path.exists(path):
        files = os.listdir(path)
    else:
        files = []
    return files


def test_clear_server():
    clrpool = 'test_clear'
    cpath = os.path.join(poolpath, clrpool)
    if not os.path.exists(cpath):
        os.makedirs(cpath)
    assert os.path.exists(cpath)
    with open(cpath+'/foo', 'w') as f:
        f.write('k')
    clear_server_poolpath(clrpool)
    assert not os.path.exists(cpath)


def populate_server(poolid):
    creators = ['Todds', 'Cassandra', 'Jane', 'Owen', 'Julian', 'Maurice']
    instruments = ['fatman', 'herscherl', 'NASA', 'CNSC', 'SVOM']

    urns = []
    for index, i in enumerate(creators):
        x = Product(description='desc ' + str(index),
                    instrument=random.choice(instruments))
        x.creator = i
        data = serialize(x)
        url = api_baseurl + poolid + '/' + prodt + '/' + str(index)
        x = requests.post(url, auth=HTTPBasicAuth(
            auth_user, auth_pass), data=data)
        o = deserialize(x.text)
        check_response(o)
        urns.append(o['result'])
    return creators, instruments, urns


def test_CRUD_product():
    ''' test saving, read, delete products API, products will be saved at /data/pool_id
    '''

    logger.info('save products')
    post_poolid = test_poolid
    clear_server_poolpath(post_poolid)

    files = [f for f in get_files(post_poolid) if f[-1].isnumeric()]
    origin_prod = len(files)

    creators, instruments, urns = populate_server(post_poolid)

    files1 = [f for f in get_files(post_poolid) if f[-1].isnumeric()]
    num_prod = len(files1)
    assert num_prod == len(creators) + origin_prod, 'Products number not match'

    newfiles = set(files1) - set(files)
    us = set(u.split(':', 2)[2].replace(':', '_') for u in urns)
    assert newfiles == us, str(newfiles) + str(us)

    # ==========
    logger.info('read product')

    u = random.choice(urns)
    # remove the leading 'urn:'
    url = api_baseurl + u[4:].replace(':', '/')
    x = requests.get(url, auth=HTTPBasicAuth(auth_user, auth_pass))
    o = deserialize(x.text)
    check_response(o)
    assert o['result'].creator == creators[urns.index(u)], 'Creator not match'

    # ===========
    ''' Test read hk api
    '''
    logger.info('read hk')
    hkpath = '/hk'
    url = api_baseurl + post_poolid + hkpath
    x = requests.get(url, auth=HTTPBasicAuth(auth_user, auth_pass))
    o = deserialize(x.text)
    check_response(o)
    assert o['result']['classes'] is not None, 'Classes jsn read failed'
    assert o['result']['tags'] is not None, 'Tags jsn read failed'
    assert o['result']['urns'] is not None, 'Urns jsn read failed'

    l = len(urns)
    inds = [int(u.rsplit(':', 1)[1]) for u in urns]
    # the last l sn's
    assert o['result']['classes'][prodt]['sn'][-l:] == inds
    assert o['result']['classes'][prodt]['currentSN'] == inds[-1]
    assert len(o['result']['tags']) == 0
    assert [d['meta']['creator'].value for
            d in list(o['result']['urns'].values())[-l:]] == creators

    logger.info('read classes')
    hkpath = '/hk/classes'
    url = api_baseurl + post_poolid + hkpath
    x = requests.get(url, auth=HTTPBasicAuth(auth_user, auth_pass))
    o = deserialize(x.text)
    check_response(o)
    assert o['result'][prodt]['sn'][-l:] == inds
    assert o['result'][prodt]['currentSN'] == inds[-1]

    logger.info('read tags')
    hkpath = '/hk/tags'
    url = api_baseurl + post_poolid + hkpath
    x = requests.get(url, auth=HTTPBasicAuth(auth_user, auth_pass))
    o = deserialize(x.text)
    check_response(o)
    assert len(o['result']) == 0

    logger.info('read urns')
    hkpath = '/hk/urns'
    url = api_baseurl + post_poolid + hkpath
    x = requests.get(url, auth=HTTPBasicAuth(auth_user, auth_pass))
    o = deserialize(x.text)
    check_response(o)

    clst = [d['meta']['creator'].value for d in list(
        o['result'].values())[-l:]]
    assert clst == creators

    # ========
    logger.info('delete a product')

    files = [f for f in get_files(post_poolid) if f[-1].isnumeric()]
    origin_prod = len(files)

    index = files[-1].rsplit('_', 1)[1]
    url = api_baseurl + post_poolid + '/fdi.dataset.product.Product/' + index
    x = requests.delete(url, auth=HTTPBasicAuth(auth_user, auth_pass))
    o = deserialize(x.text)
    check_response(o)

    files1 = [f for f in get_files(post_poolid) if f[-1].isnumeric()]
    num_prod = len(files1)
    assert num_prod + 1 == origin_prod, 'Products number not match'

    newfiles = set(files) - set(files1)
    assert len(newfiles) == 1
    f = newfiles.pop()
    assert f.endswith(str(index))

    # ========
    logger.info('delete a pool')
    files = get_files(post_poolid)
    assert len(files) != 0, 'Pool is already empty: ' + post_poolid

    url = api_baseurl + post_poolid
    x = requests.delete(url, auth=HTTPBasicAuth(auth_user, auth_pass))
    o = deserialize(x.text)
    check_response(o)

    files = get_files(post_poolid)
    assert len(files) == 0, 'Wipe pool failed: ' + o['msg']


async def lock_pool(poolid, sec):
    ''' Lock a pool from reading and return a fake response
    '''
    logger.info('Keeping files locked for %f sec' % sec)
    ppath = os.path.join(poolpath, poolid)
    # lock to prevent writing
    lock = '/tmp/fdi_locks/' + ppath.replace('/', '_') + '.write'
    logger.debug(lock)
    with filelock.FileLock(lock):
        await asyncio.sleep(sec)
    fakeres = '{"result": "FAILED", "msg": "This is a fake responses", "timestamp": ' + \
        str(time.time()) + '}'
    return deserialize(fakeres)


async def read_product(poolid):
    # trying to read
    if 1:
        prodpath = '/'+prodt+'/0'
        url = api_baseurl + poolid + prodpath
    else:
        hkpath = '/hk/classes'
        url = api_baseurl + poolid + hkpath
    logger.debug('Reading a locked file '+url)
    async with aiohttp.ClientSession() as session:
        async with session.get(url, auth=aiohttp.BasicAuth(auth_user, auth_pass)) as res:
            x = await res.text()
            o = deserialize(x)
    logger.debug("@@@@@@@locked file read: " + lls(x, 200))
    return o


def test_lock_file():
    ''' Test if a pool is locked, others can not manipulate this pool anymore before it's released
    '''
    logger.info('Test read a locked file, it will return FAILED')
    poolid = test_poolid
    # init server
    populate_server(poolid)
    #hkpath = '/hk/classes'
    #url = api_baseurl + poolid + hkpath
    #x = requests.get(url, auth=HTTPBasicAuth(auth_user, auth_pass))

    try:
        loop = asyncio.get_event_loop()
        tasks = [asyncio.ensure_future(
            lock_pool(poolid, 2)), asyncio.ensure_future(read_product(poolid))]
        taskres = loop.run_until_complete(asyncio.wait(tasks))
        loop.close()
    except Exception as e:
        logger.error('unable to start thread ' + str(e) + trbk(e))
        raise
    res = [f.result() for f in [x for x in taskres][0]]
    logger.debug('res ' + lls(res[0], 200) + '************' + lls(res[1], 200))
    if issubclass(res[0].__class__, Mapping) and 'result' in res[0] and issubclass(res[0]['result'].__class__, str):
        r1, r2 = res[0], res[1]
    else:
        r2, r1 = res[0], res[1]
    check_response(r1, True)


def test_read_non_exists_pool():
    ''' Test read a pool which doesn't exist, returns FAILED
    '''
    logger.info('Test query a pool non exist.')
    wrong_poolid = 'abc'
    prodpath = '/' + prodt + '/0'
    url = api_baseurl + wrong_poolid + prodpath
    x = requests.get(url, auth=HTTPBasicAuth(auth_user, auth_pass))
    o = deserialize(x.text)
    check_response(o, True)


def XXXtest_subclasses_pool():
    logger.info('Test create a pool which has subclass')
    poolid_1 = 'subclasses/a'
    poolid_2 = 'subclasses/b'
    prodpath = '/' + prodt + '/0'
    url1 = api_baseurl + poolid_1 + prodpath
    url2 = api_baseurl + poolid_2 + prodpath
    x = Product(description="product example with several datasets",
                instrument="Crystal-Ball", modelName="Mk II")
    data = serialize(x)
    res1 = requests.post(url1, auth=HTTPBasicAuth(
        auth_user, auth_pass), data=data)
    res2 = requests.post(url2, auth=HTTPBasicAuth(
        auth_user, auth_pass), data=data)
    o1 = deserialize(res1.text)
    o2 = deserialize(res2.text)
    check_response(o1)
    check_response(o2)

    # Wipe these pools
    url1 = api_baseurl + poolid_1
    url2 = api_baseurl + poolid_2

    res1 = requests.delete(url1,  auth=HTTPBasicAuth(auth_user, auth_pass))
    res2 = requests.delete(url2,  auth=HTTPBasicAuth(auth_user, auth_pass))
    o1 = deserialize(res1.text)
    check_response(o1)
    o2 = deserialize(res2.text)
    check_response(o2)


if __name__ == '__main__':
    now = time.time()
    node, verbose = opt(pc['node'])
    if verbose:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)
    logger.info('logging level %d' % (logger.getEffectiveLevel()))

    t = 8

    if t == 7:
        # test_lock()
        # asyncio.AbstractEventLoop.set_debug()
        loop = asyncio.get_event_loop()
        tasks = [asyncio.ensure_future(napa(5, 0)),
                 asyncio.ensure_future(napa(0.5, 0.5))]
        res = loop.run_until_complete(asyncio.wait(tasks))
        loop.close()
        print(res)

    elif t == 3:
        # test_getpnsconfig()
        test_puttestinit()
        test_putinit()
        test_getinit()
        test_getrun()
        test_putconfigpns()
        test_post()
        test_testrun()
        test_deleteclean()
        test_mirror()
        test_sleep()
    elif t == 4:
        test_serverinit()
        test_servertestinit()
        test_servertestrun()
        test_serversleep()
    elif t == 6:
        test_vvpp()

    print('test successful ' + str(time.time() - now))



# -*- coding: utf-8 -*-
import pdb
from ..pns.pnsconfig import pnsconfig as pc
from ..utils.getconfig import getConfig
from .urn import parse_poolurl

import getpass
import logging
# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))

pc.update(getConfig())

DEFAULT_MEM_POOL = 'defaultmem'
# localpool
DEFAULT_POOL = 'pool_' + getpass.getuser()


class PoolManager(object):
    """
    This class provides the means to reference ProductPool objects without having to hard-code the type of pool. For example, it could be desired to easily switch from one pool type to another.

This is done by calling the getPool() method, which will return an existing pool or create a new one if necessary.
    """
    # Global centralized dict that returns singleton -- the same -- pool for the same ID.
    _GlobalPoolList = {}
    # maps scheme to default place\poolpath
    PlacePaths = {
        'file': '/tmp',
        'mem': '/',
        'http': '127.0.0.1:5000/v0.6',
        'https': '127.0.0.1:5000/v0.6',
        'server': '/tmp/data',
    }

    @classmethod
    def getPool(cls, poolname=None, poolurl=None, **kwds):
        """ returns an instance of pool according to name or path of the pool.

        Returns the pool object if the pool is registered. Creates the pool if it does not already exist. the same poolname-path always get the same pool.

        poolname: name of the pool.
        poolurl: if given the poolpath, scheme, place will be derived from it. if not given, PoolManager.PlacePaths[scheme] is used to get poolplace and poolpath, with scheme set to 'file'. 
If poolname is missing it is derived from poolurl; if poolurl is also absent, DEFAULT_POOL is assumed.
        kwds: passed to pool instanciation arg-list.

        """
        # logger.debug('GPL ' + str(id(cls._GlobalPoolList)) +
        #             str(cls._GlobalPoolList) + ' PConf ' + str(cls.PlacePaths))

        poolpath = None
        if poolname is None:
            if poolurl:
                # the last segment will be the poolname
                poolpath, schm, pl, poolname = parse_poolurl(poolurl)
            else:
                schm = 'file'
                poolname = DEFAULT_POOL
                poolpath = cls.PlacePaths[schm]
                poolurl = schm + '://' + poolpath + '/' + poolname
                logger.debug('DEFAULT_MEM_POOL assumed: %s.' % poolurl)
            if cls.isLoaded(poolname):
                return cls._GlobalPoolList[poolname]
        else:
            if cls.isLoaded(poolname):
                return cls._GlobalPoolList[poolname]
            if poolurl:
                poolpath, schm, pl, pn = parse_poolurl(poolurl)
            else:
                schm = 'file'
                poolpath = cls.PlacePaths[schm]
                poolurl = schm + '://' + poolpath + '/' + poolname
                logger.debug('PlacePaths is used: %s.' % poolurl)

        if schm == 'file':
            from . import localpool
            p = localpool.LocalPool(
                poolname=poolname, poolurl=poolurl, **kwds)
        elif schm == 'mem':
            from . import mempool
            p = mempool.MemPool(poolname=poolname, poolurl=poolurl, **kwds)
        elif schm == 'server':
            from . import httppool
            p = httppool.HttpPool(
                poolname=poolname, poolurl=poolurl, **kwds)
        elif schm in ('http', 'https'):
            from . import httpclientpool
            p = httpclientpool.HttpClientPool(
                poolname=poolname, poolurl=poolurl, **kwds)
        else:
            raise NotImplementedError(schm + ':// is not supported')
        cls.save(poolname, p)
        logger.debug('made pool ' + str(p))
        return p

    @ classmethod
    def getMap(cls):
        """
        Returns a poolname - poolobject map.
        """
        return cls._GlobalPoolList

    @ classmethod
    def isLoaded(cls, poolname):
        """
        Whether an item with the given id has been loaded (cached).
        """
        return poolname in cls._GlobalPoolList

    @ classmethod
    def removeAll(cls):
        """ deletes all pools from the pool list, pools unwiped
        """

        cls._GlobalPoolList.clear()

    @ classmethod
    def save(cls, poolname, poolobj):
        """
        """
        cls._GlobalPoolList[poolname] = poolobj

    @ classmethod
    def getPoolurlMap(cls):
        """
        Gives the default poolurls of PoolManager.
        """
        return cls.PlacePaths

    @ classmethod
    def setPoolurlMap(cls, new):
        """
        Sets the default poolurls of PoolManager.
        """
        cls.PlacePaths.clear()
        cls.PlacePaths.update(new)

    @ classmethod
    def size(cls):
        """
        Gives the number of entries in this manager.
        """
        return len(cls._GlobalPoolList)

    def items(self):
        """
        Returns map's items
        """
        return self._GlobalPoolList.items()

    def __setitem__(self, *args, **kwargs):
        """ sets value at key.
        """
        self._GlobalPoolList.__setitem__(*args, **kwargs)

    def __getitem__(self, *args, **kwargs):
        """ returns value at key.
        """
        return self._GlobalPoolList.__getitem__(*args, **kwargs)

    def __delitem__(self, *args, **kwargs):
        """ removes value and its key.
        """
        self._GlobalPoolList.__delitem__(*args, **kwargs)

    def __len__(self, *args, **kwargs):
        """ size of data
        """
        return self._GlobalPoolList.__len__(*args, **kwargs)

    def __iter__(self, *args, **kwargs):
        """ returns an iterator
        """
        return self._GlobalPoolList.__iter__(*args, **kwargs)

    def __repr__(self):
        return self.__class__.__name__ + str(cls._GlobalPoolList)
